/**
 * <copyright>
 *
 * Copyright (c) 2002 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Contributors:
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * plugins/org.eclipse.emf.ecore.xmi/src/org/eclipse/emf/ecore/xmi/impl/XMLSaveImpl.java, emf.ecore.xmi, org.eclipse.dev, 20030620_1105VL
 * @version 1.14 6/20/03
 */
package org.eclipse.emf.ecore.xmi.impl;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.util.InternalEList;
import org.eclipse.emf.ecore.xmi.DanglingHREFException;
import org.eclipse.emf.ecore.xmi.XMIResource;
import org.eclipse.emf.ecore.xmi.XMLHelper;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xmi.XMLSave;

/**
 * This implements the XML serializer, possibly using an XMLMap
 * if one is provided as a save option.
 */
public class XMLSaveImpl implements XMLSave
{
  protected XMLHelper helper;
  protected XMLString doc;
  protected boolean declareXSI;
  protected boolean useEncodedAttributeStyle;
  protected boolean declareXML;
  protected Escape escape;
  protected Lookup featureTable;
  protected String encoding;
  protected String idAttributeName = "id";
  protected String processDanglingHREF;
  protected boolean declareSchemaLocation;

  protected static final int SKIP = 0;
  protected static final int SAME_DOC = 1;
  protected static final int CROSS_DOC = 2;

  protected static final int TRANSIENT                           = 0;
  protected static final int DATATYPE_SINGLE                     = 1;
  protected static final int DATATYPE_ELEMENT_SINGLE             = 2;
  protected static final int DATATYPE_CONTENT_SINGLE             = 3;
  protected static final int DATATYPE_SINGLE_NILLABLE            = 4;
  protected static final int DATATYPE_MANY                       = 5;
  protected static final int OBJECT_CONTAIN_SINGLE               = 6;
  protected static final int OBJECT_CONTAIN_MANY                 = 7;
  protected static final int OBJECT_HREF_SINGLE                  = 8;
  protected static final int OBJECT_HREF_MANY                    = 9;
  protected static final int OBJECT_CONTAIN_SINGLE_UNSETTABLE    = 10;
  protected static final int OBJECT_CONTAIN_MANY_UNSETTABLE      = 11;
  protected static final int OBJECT_HREF_SINGLE_UNSETTABLE       = 12;
  protected static final int OBJECT_HREF_MANY_UNSETTABLE         = 13;
  protected static final int OBJECT_ELEMENT_SINGLE               = 14;
  protected static final int OBJECT_ELEMENT_MANY                 = 15;

  protected static final String XML_VERSION = "1.0";

  protected static final String XSI_NIL             = XMIResource.XSI_NS+":"+XMIResource.NIL;             // xsi:nil
  protected static final String XSI_TYPE_NS         = XMIResource.XSI_NS+":"+XMIResource.TYPE;            // xsi:type
  protected static final String XSI_XMLNS           = XMIResource.XML_NS+":"+XMIResource.XSI_NS;          // xmlns:xsi
  protected static final String XSI_SCHEMA_LOCATION = XMIResource.XSI_NS+":"+XMIResource.SCHEMA_LOCATION; // xsi:schemaLocation

  protected static final int EMPTY_ELEMENT = 1;
  protected static final int CONTENT_ELEMENT = 2;

  public XMLSaveImpl(XMLHelper helper)
  {
    this.helper = helper;
  }

  /**
   * Constructor for XMLSave.
   * @param options
   * @param helper
   * @param encoding
   */
  public XMLSaveImpl(Map options, XMLHelper helper, String encoding)
  {
    this.helper = helper;
    init(helper.getResource(), options);
    this.encoding = encoding;
  }

  public void save(XMLResource resource, OutputStream outputStream, Map options) throws IOException
  {
    init(resource, options);
    traverse(resource.getContents());

    if (encoding.equals("US-ASCII") || encoding.equals("ASCII"))
    {
      writeAscii(outputStream);
      outputStream.flush();
    }
    else
    {
      OutputStreamWriter outputStreamWriter = new OutputStreamWriter(outputStream, helper.getJavaEncoding(encoding));
      write(outputStreamWriter);
      outputStreamWriter.flush();
    }

    featureTable = null;
    doc = null;

    if (processDanglingHREF == null ||
        XMLResource.OPTION_PROCESS_DANGLING_HREF_THROW.equals(processDanglingHREF))
    {
      DanglingHREFException exception = helper.getDanglingHREFException();

      if (exception != null)
      {
        helper = null;
        throw new Resource.IOWrappedException(exception);
      }
    }

    helper = null;
  }

  protected void init(XMLResource resource, Map options)
  {
    declareXSI = false;
    useEncodedAttributeStyle = Boolean.TRUE.equals(options.get(XMLResource.OPTION_USE_ENCODED_ATTRIBUTE_STYLE));
    declareXML = !Boolean.FALSE.equals(options.get(XMLResource.OPTION_DECLARE_XML));
    declareSchemaLocation = Boolean.TRUE.equals(options.get(XMLResource.OPTION_SCHEMA_LOCATION));
    Integer lineWidth = (Integer)options.get(XMLResource.OPTION_LINE_WIDTH);
    doc = new XMLString(lineWidth == null ? Integer.MAX_VALUE : lineWidth.intValue());
    escape =
      Boolean.TRUE.equals(options.get(XMLResource.OPTION_SKIP_ESCAPE)) ?
        null :
        new Escape();

    if (options.containsKey(XMLResource.OPTION_ENCODING))
    {
      encoding = (String)options.get(XMLResource.OPTION_ENCODING);
    }
    else if (resource != null)
    {
      encoding = resource.getEncoding();
    }

    processDanglingHREF = (String) options.get(XMLResource.OPTION_PROCESS_DANGLING_HREF);
    helper.setProcessDanglingHREF(processDanglingHREF);

    XMLResource.XMLMap map = (XMLResource.XMLMap) options.get(XMLResource.OPTION_XML_MAP);

    if (map != null)
    {
      helper.setXMLMap(map);

      if (map.getIDAttributeName() != null)
      {
        idAttributeName = map.getIDAttributeName();
      }
    }

    featureTable = new Lookup(map);
  }

  public void traverse(List contents)
  {
    if (declareXML)
    {
      doc.add("<?xml version=\"" + XML_VERSION + "\" encoding=\"" + encoding + "\"?>");
      doc.addLine();
    }

    int size = contents.size();

    // Reserve a place to insert xmlns declarations after we know what they all are.
    //
    Object mark;

    if (size == 1)
    {
      mark = writeTopObject((EObject) contents.get(0));
    }
    else
    {
      mark = writeTopObjects(contents);
    }

    // Go back and add all the XMLNS stuff.
    //
    doc.resetToMark(mark);
    addNamespaceDeclarations();
  }

  protected Object writeTopObject(EObject top)
  {
    EClass eClass = top.eClass();
    String name = helper.getQName(eClass);
    doc.startElement(name);
    Object mark = doc.mark();
    saveElementID(top);
    return mark;
  }

  protected Object writeTopObjects(List contents)
  {
    EObject top = (EObject) contents.get(0);
    EClass eClass = top.eClass();

    String name = helper.getQName(eClass);
    doc.startElement(name);
    Object mark = doc.mark();
    saveElementID(top);

    return mark;
  }

  protected void addNamespaceDeclarations()
  {
    EPackage[] packages = helper.packages();
    StringBuffer xsiSchemaLocation = null;
    if (declareSchemaLocation)
    {
      for (int i = 0; i < packages.length; i++)
      {
        EPackage ePackage = packages[i];

        EObject root = EcoreUtil.getRootContainer(ePackage);
        if (root instanceof EPackage)
        {
          EPackage rootEPackage = (EPackage)root;
          Resource resource = rootEPackage.eResource();
          if (resource != null)
          {
            URI uri = resource.getURI();
            String rootNsURI = rootEPackage.getNsURI();
            if (uri == null ? rootNsURI != null : !uri.toString().equals(rootNsURI))
            {
              declareXSI = true;
              if (xsiSchemaLocation == null)
              {
                xsiSchemaLocation = new StringBuffer();
              }
              else
              {
                xsiSchemaLocation.append(' ');
              }
              xsiSchemaLocation.append(rootNsURI);
              xsiSchemaLocation.append(' ');
              xsiSchemaLocation.append(helper.getHREF(rootEPackage));
            }
          }
        }
      }
    }

    if (declareXSI)
    {
      doc.addAttribute(XSI_XMLNS, XMIResource.XSI_URI);
    }

    for (int i = 0; i < packages.length; i++)
    {
      EPackage ePackage = packages[i];
      String nsURI = ePackage.getNsURI();
      String nsPrefix = ePackage.getNsPrefix();
      doc.addAttributeNS(XMIResource.XML_NS, nsPrefix, nsURI);
    }

    if (xsiSchemaLocation != null)
    {
      doc.addAttribute(XSI_SCHEMA_LOCATION, xsiSchemaLocation.toString());
    }
  }

  public void write(OutputStreamWriter os) throws IOException
  {
    final int BUFFER_SIZE = 8192;
    char[] buffer = new char[BUFFER_SIZE];
    int pos = 0;
    for (Iterator i = doc.iterator(); i.hasNext(); )
    {
      String s = (String)i.next();
      int slen = s.length();
      if (slen+pos >= buffer.length)
      {
        os.write(buffer, 0, pos);
        pos = 0;
        if (slen > buffer.length) {
            buffer = new char[slen];
        }
      }
      s.getChars(0, slen, buffer, pos);
      pos += slen;
    }
    os.write(buffer, 0, pos);
    os.flush();
  }

  public void writeAscii(OutputStream os) throws IOException
  {
    final int BUFFER_SIZE = 8192;
    char[] buffer = new char[BUFFER_SIZE];
    byte[] bytes = new byte[BUFFER_SIZE];
    int pos = 0;
    for (Iterator i = doc.iterator(); i.hasNext(); )
    {
      String s = (String) i.next();
      int slen = s.length();
      if (slen + pos >= buffer.length)
      {
        for (int x = 0; x < pos; x++)
        {
          bytes[x] = (byte) (buffer[x] & 0xFF);
        }
        os.write(bytes, 0, pos);
        pos = 0;
        if (slen > buffer.length)
        {
          buffer = new char[slen];
          bytes = new byte[slen];
        }
      }
      s.getChars(0, slen, buffer, pos);
      pos += slen;
    }

    for (int x = 0; x < pos; x++)
    {
      bytes[x] = (byte) (buffer[x] & 0xFF);
    }

    os.write(bytes, 0, pos);
    os.flush();
  }

  public char[] toChar()
  {
    int size = doc.getLength();
    char[] output = new char[size];
    doc.getChars(output, 0);
    return output;
  }

  protected void saveElement(EObject o, EStructuralFeature f)
  {
    XMLResource.XMLMap map = ((XMLHelper) helper).getXMLMap();
    EClass eClass = o.eClass();
    XMLResource.XMLInfo info = map == null ? null : (XMLResource.XMLInfo)map.getInfo(eClass);
    boolean checkXSIType = true;
    if (info != null && info.getXMLRepresentation() == XMLResource.XMLInfo.ELEMENT)
    {
      String elementName = helper.getQName(eClass);
      doc.startElement(elementName);
      checkXSIType = false;
    }
    else
    {
      String featureName = helper.getQName(f);
      doc.startElement(featureName);
    }

    if (checkXSIType && eClass != f.getEType())
    {
      saveTypeAttribute(eClass);
    }

    saveElementID(o);
  }

  protected void saveTypeAttribute(EClass eClass)
  {
    declareXSI = true;
    doc.addAttribute(XSI_TYPE_NS, helper.getQName(eClass));
  }

  protected boolean saveFeatures(EObject o)
  {
    EClass eClass = o.eClass();
    EStructuralFeature[] features = featureTable.getFeatures(eClass);
    int[] featureKinds = featureTable.getKinds(eClass, features);
    int[] elementFeatures = null;
    int elementCount = 0;

    // Process XML attributes
    for (int i = 0; i < features.length; i++ )
    {
      int kind = featureKinds[i];
      EStructuralFeature f = features[i];
      if (kind != TRANSIENT && o.eIsSet(f))
      {
        switch (kind)
        {
          case DATATYPE_ELEMENT_SINGLE:
          {
            if (elementFeatures == null)
            {
              elementFeatures = new int[features.length];
            }
            elementFeatures[elementCount++] = i;
            break;
          }
          case DATATYPE_SINGLE:
          {
            saveDataTypeSingle(o, f);
            break;
          }
          case DATATYPE_SINGLE_NILLABLE:
          {
            if (isNil(o, f))
            {
              if (elementFeatures == null)
              {
                elementFeatures = new int[features.length];
              }
              elementFeatures[elementCount++] = i;
            }
            else
            {
              saveDataTypeSingle(o, f);
            }
            break;
          }
          case OBJECT_HREF_SINGLE_UNSETTABLE:
          {
            if (isNil(o, f))
            {
              if (elementFeatures == null)
              {
                elementFeatures = new int[features.length];
              }
              elementFeatures[elementCount++] = i;
              break;
            }
            // it's intentional to keep going
          }
          case OBJECT_HREF_SINGLE:
          {
            if (useEncodedAttributeStyle)
            {
              saveEObjectSingle(o, f);
            }
            else
            {
              switch (sameDocSingle(o, f))
              {
                case SAME_DOC:
                {
                  saveIDRefSingle(o, f);
                  break;
                }
                case CROSS_DOC:
                {
                  if (elementFeatures == null)
                  {
                    elementFeatures = new int[features.length];
                  }
                  elementFeatures[elementCount++] = i;
                  break;
                }
              }
            }
            break;
          }
          case OBJECT_HREF_MANY_UNSETTABLE:
          {
            if (isEmpty(o, f))
            {
              saveManyEmpty(f);
              break;
            }
            // it's intentional to keep going
          }
          case OBJECT_HREF_MANY:
          {
            if (useEncodedAttributeStyle)
            {
              saveEObjectMany(o, f);
            }
            else
            {
              switch (sameDocMany(o, f))
              {
                case SAME_DOC:
                {
                  saveIDRefMany(o, f);
                  break;
                }
                case CROSS_DOC:
                {
                  if (elementFeatures == null)
                  {
                    elementFeatures = new int[features.length];
                  }
                  elementFeatures[elementCount++] = i;
                  break;
                }
              }
            }
            break;
          }
          case OBJECT_CONTAIN_MANY_UNSETTABLE:
          case DATATYPE_MANY:
          {
            if (isEmpty(o, f))
            {
              saveManyEmpty(f);
              break;
            }
            // it's intentional to keep going
          }
          case OBJECT_CONTAIN_SINGLE_UNSETTABLE:
          case OBJECT_CONTAIN_SINGLE:
          case OBJECT_CONTAIN_MANY:
          case OBJECT_ELEMENT_SINGLE:
          case OBJECT_ELEMENT_MANY:
          {
            if (elementFeatures == null)
            {
              elementFeatures = new int[features.length];
            }
            elementFeatures[elementCount++] = i;
            break;
          }
        }
      }
    }

    String content = getContent(o, features);

    if (elementFeatures == null)
    {
      if (content == null)
      {
        endSaveFeatures(o, EMPTY_ELEMENT, null);
        return false;
      }
      else
      {
        endSaveFeatures(o, CONTENT_ELEMENT, content);
        return true;
      }
    }

    // Process XML elements
    for (int i = 0; i < elementCount; i++ )
    {
      int kind = featureKinds[elementFeatures[i]];
      EStructuralFeature f = features[elementFeatures[i]];
      switch (kind)
      {
        case DATATYPE_SINGLE_NILLABLE:
        {
          saveNil(f);
          break;
        }
        case DATATYPE_MANY:
        {
          saveDataTypeMany(o, f);
          break;
        }
        case DATATYPE_ELEMENT_SINGLE:
        {
          saveDataTypeElementSingle(o, f);
          break;
        }
        case OBJECT_CONTAIN_SINGLE_UNSETTABLE:
        {
          if (isNil(o, f))
          {
            saveNil(f);
            break;
          }
          // it's intentional to keep going
        }
        case OBJECT_CONTAIN_SINGLE:
        {
          saveContainedSingle(o, f);
          break;
        }
        case OBJECT_CONTAIN_MANY_UNSETTABLE:
        case OBJECT_CONTAIN_MANY:
        {
          saveContainedMany(o, f);
          break;
        }
        case OBJECT_HREF_SINGLE_UNSETTABLE:
        {
          if (isNil(o, f))
          {
            saveNil(f);
            break;
          }
          // it's intentional to keep going
        }
        case OBJECT_HREF_SINGLE:
        {
          saveHRefSingle(o, f);
          break;
        }
        case OBJECT_HREF_MANY_UNSETTABLE:
        case OBJECT_HREF_MANY:
        {
          saveHRefMany(o, f);
          break;
        }
        case OBJECT_ELEMENT_SINGLE:
        {
          saveElementReferenceSingle(o, f);
          break;
        }
        case OBJECT_ELEMENT_MANY:
        {
          saveElementReferenceMany(o, f);
          break;
        }
      }
    }
    endSaveFeatures(o, 0, null);
    return true;
  }

  protected void endSaveFeatures(EObject o, int elementType, String content)
  {
    switch (elementType)
    {
      case EMPTY_ELEMENT:
      {
        doc.endEmptyElement();
        break;
      }
      case CONTENT_ELEMENT:
      {
        doc.endContentElement(content);
        break;
      }
      default:
      {
        doc.endElement();
        break;
      }
    }
  }

  protected void saveDataTypeSingle(EObject o, EStructuralFeature f)
  {
    EDataType d = (EDataType)f.getEType();
    EPackage ePackage = d.getEPackage();
    EFactory fac = ePackage.getEFactoryInstance();
    Object value = helper.getValue(o, f);
    if (value != null)
    {
      String svalue = fac.convertToString(d, value);
      if (escape != null)
      {
        svalue = escape.convert(svalue);
      }
      doc.addAttribute(helper.getQName(f), svalue);
    }
  }

  protected boolean isNil(EObject o, EStructuralFeature f)
  {
    return helper.getValue(o, f) == null;
  }

  protected boolean isEmpty(EObject o, EStructuralFeature f)
  {
    return ((List)helper.getValue(o, f)).isEmpty();
  }

  protected void saveNil(EStructuralFeature f)
  {
    String name = helper.getQName(f);
    doc.startElement(name);
    doc.addAttribute(XSI_NIL, "true");
    doc.endEmptyElement();
    declareXSI = true;
  }

  protected void saveManyEmpty(EStructuralFeature f)
  {
    doc.addAttribute(helper.getQName(f), "");
  }

  protected void saveDataTypeMany(EObject o, EStructuralFeature f)
  {
    EDataType d = (EDataType)f.getEType();
    EPackage ePackage = d.getEPackage();
    EFactory fac = ePackage.getEFactoryInstance();

    List values = (List)helper.getValue(o, f);
    int size = values.size();
    if (size > 0)
    {
      String name = helper.getQName(f);
      for (int i = 0; i < size; ++i)
      {
        Object value = values.get(i);
        if (value == null)
        {
          doc.startElement(name);
          doc.addAttribute(XSI_NIL, "true");
          doc.endEmptyElement();
          declareXSI = true;
        }
        else
        {
          doc.startElement(name);
          String svalue = fac.convertToString(d, value);
          if (escape != null)
          {
            svalue = escape.convert(svalue);
          }
          doc.endContentElement(svalue);
        }
      }
    }
  }

  protected void saveEObjectSingle(EObject o, EStructuralFeature f)
  {
    EObject value = (EObject)helper.getValue(o, f);
    if (value != null)
    {
      String name = helper.getQName(f);
      String id = helper.getHREF(value);
      if (id != null)
      {
        doc.startAttribute(name);
        if (!id.startsWith("#"))
        {
          EClass eClass = value.eClass();
          EClass expectedType = (EClass) f.getEType();
          if (eClass != expectedType && expectedType.isAbstract())
          {
            doc.addAttributeContent(helper.getQName(eClass));
            doc.addAttributeContent(" ");
          }
        }
        doc.addAttributeContent(id);
        doc.endAttribute();
      }
    }
  }

  protected void saveEObjectMany(EObject o, EStructuralFeature f)
  {
    InternalEList values = (InternalEList)helper.getValue(o, f);

    if (!values.isEmpty())
    {
      String name = helper.getQName(f);
      doc.startAttribute(name);

      for (Iterator i = values.basicIterator(); ; )
      {
        EObject value = (EObject)i.next();
        String id = helper.getHREF(value);
        if (id != null)
        {
          if (!id.startsWith("#"))
          {
            EClass eClass = value.eClass();
            EClass expectedType = (EClass) f.getEType();
            if (eClass != expectedType && expectedType.isAbstract())
            {
              doc.addAttributeContent(helper.getQName(eClass));
              doc.addAttributeContent(" ");
            }
          }
          doc.addAttributeContent(id);
        }
        if (i.hasNext())
        {
          doc.addAttributeContent(" ");
        }
        else
        {
          break;
        }
      }
      doc.endAttribute();
    }
  }

  protected void saveIDRefSingle(EObject o, EStructuralFeature f)
  {
    EObject value = (EObject)helper.getValue(o, f);
    if (value != null)
    {
      String name = helper.getQName(f);
      String id = helper.getIDREF(value);
      doc.addAttribute(name, id);
    }
  }

  protected void saveIDRefMany(EObject o, EStructuralFeature f)
  {
    InternalEList values = (InternalEList)helper.getValue(o, f);
    if (!values.isEmpty())
    {
      String name = helper.getQName(f);
      StringBuffer ids = new StringBuffer(values.size() * 10);
      for (Iterator i = values.basicIterator();;)
      {
        EObject value = (EObject)i.next();
        String id = helper.getIDREF(value);
        ids.append(id);
        if (i.hasNext())
        {
          ids.append(" ");
        }
        else
        {
          break;
        }
      }
      doc.addAttribute(name, ids.toString());
    }
  }

  protected void saveElementReference(EObject remote, EStructuralFeature f)
  {
    String name = helper.getQName(f);
    String href = helper.getHREF(remote);
    if (href != null)
    {
      doc.startElement(name);
      doc.endContentElement(href);
    }
  }

  protected void saveElementReferenceSingle(EObject o, EStructuralFeature f)
  {
    EObject value = (EObject)helper.getValue(o, f);
    if (value != null)
    {
      saveElementReference(value, f);
    }
  }

  protected void saveElementReferenceMany(EObject o, EStructuralFeature f)
  {
    InternalEList values = (InternalEList)helper.getValue(o, f);
    int size = values.size();
    for (int i = 0; i < size; i++)
    {
      saveElementReference((EObject)values.get(i), f);
    }
  }

  protected void saveHref(EObject remote, EStructuralFeature f)
  {
    String name = helper.getQName(f);
    String href = helper.getHREF(remote);
    if (href != null)
    {
      doc.startElement(name);
      EClass eClass = remote.eClass();

      EClass expectedType = (EClass) f.getEType();
      if (eClass != expectedType && expectedType.isAbstract())
      {
        saveTypeAttribute(eClass);
      }

      doc.addAttribute(XMIResource.HREF, href);
      doc.endEmptyElement();
    }
  }

  protected void saveHRefSingle(EObject o, EStructuralFeature f)
  {
    EObject value = (EObject)helper.getValue(o, f);
    if (value != null)
    {
      saveHref(value, f);
    }
  }

  protected void saveHRefMany(EObject o, EStructuralFeature f)
  {
    InternalEList values = (InternalEList)helper.getValue(o, f);
    int size = values.size();
    for (int i = 0; i < size; i++)
    {
      saveHref((EObject)values.get(i), f);
    }
  }

  protected void saveContainedSingle(EObject o, EStructuralFeature f)
  {
    EObject value = (EObject)helper.getValue(o, f);
    if (value != null)
    {
      saveElement(value, f);
    }
  }

  protected void saveContainedMany(EObject o, EStructuralFeature f)
  {
    List values = (List)helper.getValue(o, f);
    int size = values.size();
    for (int i = 0; i < size; i++)
    {
      EObject value = (EObject)values.get(i);
      if (value != null)
      {
        saveElement(value, f);
      }
    }
  }

  protected int sameDocSingle(EObject o, EStructuralFeature f)
  {
    InternalEObject value = (InternalEObject)helper.getValue(o, f);
    if (value == null)
    {
      return SKIP;
    }
    else if (value.eIsProxy())
    {
      return CROSS_DOC;
    }
    else
    {
      Resource res = value.eResource();
      return res == helper.getResource() ? SAME_DOC : CROSS_DOC;
    }
  }

  protected int sameDocMany(EObject o, EStructuralFeature f)
  {
    InternalEList values = (InternalEList)helper.getValue(o, f);
    if (values.isEmpty())
    {
      return SKIP;
    }

    for (Iterator i = values.basicIterator(); i.hasNext(); )
    {
      InternalEObject value = (InternalEObject)i.next();
      if (value.eIsProxy() || value.eResource() != helper.getResource()) {
        return CROSS_DOC;
      }
    }

    return SAME_DOC;
  }

  protected String getContent(EObject o, EStructuralFeature[] features)
  {
    XMLResource.XMLMap map = ((XMLHelper) helper).getXMLMap();

    if (map == null)
    {
      return null;
    }

    for (int i = 0; i < features.length; i++)
    {
      XMLResource.XMLInfo info = map.getInfo(features[i]);

      if (info != null && info.getXMLRepresentation() == XMLResource.XMLInfo.CONTENT)
      {
        Object value = helper.getValue(o, features[i]);
        if (value == null)
        {
          return null;
        }
        else
        {
          EDataType d = (EDataType) features[i].getEType();
          EPackage ePackage = d.getEPackage();
          EFactory fac = ePackage.getEFactoryInstance();
          String svalue = fac.convertToString(d, value);
          if (escape != null)
          {
            svalue = escape.convert(svalue);
          }
          return svalue;
        }
      }
    }

    return null;
  }

  protected void saveDataTypeElementSingle(EObject o, EStructuralFeature f)
  {
    String name = helper.getQName(f);
    Object value = helper.getValue(o, f);
    if (value == null)
    {
      doc.startElement(name);
      doc.addAttribute(XSI_NIL, "true");
      doc.endEmptyElement();
      declareXSI = true;
    }
    else
    {
      EDataType d = (EDataType)f.getEType();
      EPackage ePackage = d.getEPackage();
      EFactory fac = ePackage.getEFactoryInstance();
      doc.startElement(name);
      String svalue = fac.convertToString(d, value);

      if (escape != null)
      {
        svalue = escape.convert(svalue);
      }

      doc.endContentElement(svalue);
    }
  }

  protected void saveElementID(EObject o)
  {
    String id = helper.getID(o);
    if (id != null)
    {
      doc.addAttribute(idAttributeName, id);
    }

    saveFeatures(o);
  }

  protected static class Lookup
  {
    protected static final int SIZE = 4095;  // 2^N-1
    protected EClass[] classes;
    protected EStructuralFeature[][] features;
    protected int[][] featureKinds;
    protected XMLResource.XMLMap map;

    Lookup(XMLResource.XMLMap map)
    {
      this.map = map;
      classes = new EClass[SIZE + 1];
      features = new EStructuralFeature[SIZE + 1][];
      featureKinds = new int[SIZE + 1][];
    }

    public EStructuralFeature[] getFeatures(EClass cls)
    {
      int index = cls.hashCode() & SIZE;
      EClass c = classes[index];

      if (c == cls)
      {
        return features[index];
      }

      EStructuralFeature[] featureList = listFeatures(cls);
      if (c == null)
      {
        classes[index] = cls;
        features[index] = featureList;
        featureKinds[index] = listKinds(featureList);
      }
      return featureList;
    }

    public int[] getKinds(EClass cls, EStructuralFeature[] featureList)
    {
      int index = cls.hashCode() & SIZE;
      EClass c = classes[index];

      if (c == cls)
      {
        return featureKinds[index];
      }

      int[] kindsList = listKinds(featureList);
      if (c == null)
      {
        classes[index] = cls;
        features[index] = featureList;
        featureKinds[index] = kindsList;
      }
      return kindsList;
    }

    protected EStructuralFeature[] listFeatures(EClass cls)
    {
      List f = map == null ? cls.getEAllStructuralFeatures() : map.getFeatures(cls);
      return (EStructuralFeature[]) f.toArray(new EStructuralFeature[f.size()]);
    }

    protected int[] listKinds(EStructuralFeature[] featureList)
    {
      int[] kinds = new int[featureList.length];
      for (int i = featureList.length-1; i >= 0; i--)
      {
        kinds[i] = featureKind(featureList[i]);
      }

      return kinds;
    }

    protected int featureKind(EStructuralFeature f)
    {
      if (f.isTransient())
      {
        return TRANSIENT;
      }

      boolean isMany = f.isMany();
      boolean isUnsettable = f.isUnsettable();

      if (f instanceof EReference)
      {
        EReference r = (EReference)f;
        if (r.isContainment())
        {
          return
            isMany ?
              isUnsettable ? OBJECT_CONTAIN_MANY_UNSETTABLE : OBJECT_CONTAIN_MANY :
              isUnsettable ? OBJECT_CONTAIN_SINGLE_UNSETTABLE : OBJECT_CONTAIN_SINGLE;
        }
        EReference opposite = r.getEOpposite();
        if (opposite != null && opposite.isContainment())
        {
          return TRANSIENT;
        }

        if (map != null)
        {
          XMLResource.XMLInfo info = map.getInfo(f);
          if (info != null && info.getXMLRepresentation() == XMLResource.XMLInfo.ELEMENT)
          {
            return
              isMany ?
                OBJECT_ELEMENT_MANY : 
                OBJECT_ELEMENT_SINGLE;
          }
        }

        return
          isMany ?
            isUnsettable ? OBJECT_HREF_MANY_UNSETTABLE : OBJECT_HREF_MANY :
            isUnsettable ? OBJECT_HREF_SINGLE_UNSETTABLE : OBJECT_HREF_SINGLE;
      }
      else
      {
        // Attribute
        EDataType d = (EDataType) f.getEType();
        if (!d.isSerializable())
        {
          return TRANSIENT;
        }
        if (isMany)
        {
          return DATATYPE_MANY;
        }

        if (isUnsettable && map == null)
        {
          return DATATYPE_SINGLE_NILLABLE;
        }

        if (map == null)
        {
          return DATATYPE_SINGLE;
        }
        else
        {
          XMLResource.XMLInfo info = map.getInfo(f);

          if (info != null && info.getXMLRepresentation() == XMLResource.XMLInfo.ELEMENT)
          {
            return DATATYPE_ELEMENT_SINGLE;
          }
          else if (info != null && info.getXMLRepresentation() == XMLResource.XMLInfo.CONTENT)
          {
            return DATATYPE_CONTENT_SINGLE;
          }
          else
          {
            if (isUnsettable)
              return DATATYPE_SINGLE_NILLABLE;
            else
              return DATATYPE_SINGLE;
          }
        }
      }
    }
  }

  protected static class Escape
  {
    protected char[] value;

    protected final char[] AMP   = { '&', 'a', 'm', 'p', ';' };
    protected final char[] LESS  = { '&', 'l', 't',';' };
    protected final char[] QUOTE = { '&', 'q', 'u', 'o', 't', ';' };
    protected final char[] LF    = { '&', '#', 'x', 'A', ';' };
    protected final char[] CR    = { '&', '#', 'x', 'D', ';' };
    protected final char[] TAB   = { '&', '#', 'x', '9', ';' };

    Escape()
    {
      value = new char[100];
    }

    /*
     *  Convert:
     *  & to &amp;
     *  < to &lt;
     *  " to &quot;
     *  \t to &#x9;
     *  \n to &#xA;
     *  \r to &#xD;
     */
    public String convert(String input)
    {
      boolean changed = false;
      int inputLength = input.length();
      grow(inputLength);
      input.getChars(0, inputLength, value, 0);
      int pos = 0;
      while (inputLength-- > 0)
      {
        switch (value[pos])
        {
          case '&':
            pos = replace(pos, AMP, inputLength);
            changed = true;
            break;
          case '<':
            pos = replace(pos, LESS, inputLength);
            changed = true;
            break;
          case '"':
            pos = replace(pos, QUOTE, inputLength);
            changed = true;
            break;
          case '\n':
          {
            pos = replace(pos, LF, inputLength);
            changed = true;
            break;
          }
          case '\r':
          {
            pos = replace(pos, CR, inputLength);
            changed = true;
            break;
          }
          case '\t':
          {
            pos = replace(pos, TAB, inputLength);
            changed = true;
            break;
          }
          default:
            pos++;
            break;
        }
      }

      return changed ? new String(value, 0, pos) : input;
    }

    protected int replace(int pos, char[] replacement, int inputLength)
    {
      int rlen = replacement.length;
      int newPos = pos + rlen;
      grow(newPos + inputLength);
      System.arraycopy(value, pos+1, value, newPos, inputLength);
      System.arraycopy(replacement, 0, value, pos, rlen);
      return newPos;
    }

    protected void grow(int newSize)
    {
      int vlen = value.length;
      if (vlen < newSize)
      {
        char[] newValue = new char[newSize + newSize/2];
        System.arraycopy(value, 0, newValue, 0, vlen);
        value = newValue;
      }
    }
  }
}
